/*	inline.h

	AFAPI SHMAPERS inline primitive functions....

	Unlike other versions of AFAPI, this library is intended to
	obtain maximum performance from stock hardware.  In particular,
	this is optimized for SMP machines that support System V
	shared memory (which is generally grouped with the System V
	IPC support...  make sure you have it enabled in your kernel).
	If you are not sure, try typing "ipcs"; if you get a status
	report on shared memory, semaphores, and message queues, you
	should be fine.

	Note that, although this library will run on single-processor
	machines, it will spend a lot of time spinning that way.

	July 1996 by Hank Dietz
*/


#ifndef	_INLINE_H
#define	_INLINE_H


#include "afapi.h"

/*	p__wait()...


	There are lots of ways to barrier synchronize using shared
	memory, including the use of various atomic operations.
	However, atomic spin locks can truly swamp the bus.

	In contrast, this code uses no atomic locks.  Basically, the
	idea is that reads from a shared object can be cached, hence
	they will not cause bus traffic until another processor
	invalidates.  The use of a counter for each PE simplifies the
	code and also allows a two-phase rather than a four-phase
	barrier.  There is also the minor optimization that if you
	find any PE is waiting for the next barrier, then he must have
	already found everyone at this barrier, so I'm done
	immediately.

	The final optimization is the insertion of a P_SHM_IDLE as a
	suggestion to the UNIX scheduler that this job may be
	spin-locked for some time (i.e., schedule another job).  In
	types.h, P_SHM_IDLE should be set to usleep(1) for heavily
	loaded systems or systems with slow coherence mechanisms; for
	fast, lightly-loaded, systems, define it to be nothing.
*/

extern inline
void
p__wait(void)
{
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = ++(p_shm_me->barno);
	register int barnonext = (barno + 1);

	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) return;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) return;
			} while (hisbarno != barno);
		}
	} while (++p < q);
}

/*	p__waitbar(), p__any(), and p_all()...
*/

extern inline
p_bar
p__waitbar(register p_1u f)
{
	/* Collect the array of flags and compress it to bits... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);
	register p_bar result = 0;

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Convert value array into packed bits... */
	p = p_shm_all;
	do {
		--q;
		result = ((result << 1) | q->datum[ which ].u_1u);
	} while (q > p);

	return(result);
}


extern inline
p_1u
p__any(register p_1u f)
{
	/* Collect the array of flags and look for a 1... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for a 1... */
	p = p_shm_all;
	do {
		if (p->datum[ which ].u_1u) return(1);
	} while (++p < q);

	return(0);
}


extern inline
p_1u
p__all(register p_1u f)
{
	/* Collect the array of flags and look for a 0... */
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = p_shm_me->barno + 1;
	register int which = (barno & 1);
	register int barnonext = (barno + 1);

	/* Install my datum and then signal barrier */
	p_shm_me->datum[ which ].u_1u = (f != 0);
	p_shm_me->barno = barno;

	/* Now wait for everybody else... */
	do {
		register int hisbarno = p->barno;
		if (hisbarno == barnonext) goto syncd;
		if (hisbarno != barno) {
			do {
				P_SHM_IDLE;
				hisbarno = p->barno;
				if (hisbarno == barnonext) goto syncd;
			} while (hisbarno != barno);
		}
	} while (++p < q);

syncd:
	/* Scan for a 1... */
	p = p_shm_all;
	do {
		if (0 == p->datum[ which ].u_1u) return(0);
	} while (++p < q);

	return(1);
}

/*	Inline signal primitives...

	Signals are really very easy because we use the barrier
	number.  If somebody is at the next count, we know that
	they are sending a signal....
*/

extern inline void
s__signal(register int s)
{
	/* This is not worth inlining... */
	s_signal(s);
}


extern inline void
s__poll(void)
{
	register volatile p_shm_type *p = p_shm_all;
	register volatile p_shm_type *q = (p + NPROC);
	register int barno = (p_shm_me->barno + 1);

	/* Is anyone at the next barrier? */
	do {
		/* If so, handle the signal... */
		if (p->barno == barno) {
			s_signal(S_SIGANY);
			return;
		}
	} while (++p < q);
}


extern inline void
s__wait(void)
{
	/* Just exit and re-enter... that flushes things */
	s_exit();
	s_init();
}

/*	Inline RSM interface...

	It happens that only the signal handler itself is a lot of
	code, thus, these can also be inlined....
*/

extern inline void
s__update(register void *a,
register int n)
{
	s_updateArgs.a = a;
	s_updateArgs.n = n;
	s__signal(S_SIGMEM);
}


extern inline void
s__initLock(register s_lock *lock)
{
	/* Initialize lock... */
	*lock = S_INITLOCK;
	s__update(((void *) lock), sizeof(s_lock));
}


extern inline void
s__acquire(register s_lock *lock)
{
	/* Acquire lock...  the funky thing about this is that it
	   works for our system BECAUSE replicated shared memory
	   updates are internally executed in a purely synchronous
	   way...  it wouldn't work using the SMP hardware directly.
	*/
	register s_lock us = (IPROC + 1);

	do {
		/* Wait for lock to be open... */
		while (*lock != S_INITLOCK) {
			s__poll();
		}

		/* Try to grab it for us */
		*lock = us;
		s__update(((void *) lock), sizeof(s_lock));

		/* Continue looping if we lost the race... */
	} while (*lock != us);
}


extern inline void
s__release(register s_lock *lock)
{
	/* Unlock lock.
	   Hmm...  this looks like s__initLock(), doesn't it?
	*/
	*lock = S_INITLOCK;
	s__update(((void *) lock), sizeof(s_lock));
}


#endif
